function [W,H,D] = SISC(V, d, varargin)
% SISC: Shift Invariant Sparse Coding
%
% Authors:
%   Morten Mrup and Mikkel N. Schmidt 
%   Technical University of Denmark, 
%   Institute for Matematical Modelling
%
% Reference:
%   M. Mrup and M. N. Schmidt. and Lars K. Hansen "Shift Invariant Sparse
%   Coding of Image and Music data", submitted JMLR 2007
% 
%   Notation correspondence to the above article: 
%       W(x,d,y)=\Phi_d(x,y), H(d,u,v)=\alpha_d(u,v), D(c,d)=S(c,d)
%
% Usage:
%   [W,H,D] = SISC(V, d, varargin)
%
% Input:
%   V                 M x N x K data matrix
%   d                 Number of factors
%   options
%     .Tau            Tau(1) number of rows in basis images, Tau(2)
%                     number of columns
%     .Phi            vertical Shifts of W (default full size of image)
%     .costfcn        Cost function to optimize
%                       'ls': Least squares (default)
%                       'kl': Kullback Leibler
%     .W              Initial W, array of size Tau(1) x d x Tau(2)
%     .H              Initial H, array of size d x N x Phi
%     .D              Initial D, matrix of size C x d 
%     .Q              Indicator matrix of same size as V, 1 datapoint in V
%                     is present 0 datapoint in V is missing
%     .const_WHD      vector with 3 elements (default [0 0 0]), value of 0:  Update, 1: Don't
%                     update, i.e.
%                           const_WHD(1)=0 update W
%                           const_WHD(2)=1 don't update H
%                           const_WHD(3)=0 update D                       
%     .sparseD        0: D normalized by |D|_F, 1: D normalized by |D_k|_F 2: sparsity also imposed on D with
%                       weight lambda (default 0)
%     .lambda         Sparsity weight on H and potentially D 
%                       (sparsity function used is the 1-norm)
%     .phiAlign       1: align Phi shifts, 0: don't align Phi shifts (default 0)
%     .tauAlign       1: align Phi shifts, 0: don't align Phi shifts (default 0)
%     .maxiter        Maximum number of iterations (default 100)
%     .conv_criteria  Function exits when cost/delta_cost exceeds this
%     .accel          Wild driver accelleration parameter (default 1.3)
%     .displaylevel   Level of display: [off | final | iter]
% 
% Output:
%   W                 Tau(1) x d x Tau(2) array
%   H                 d x N x Phi array
%   D                 K x d matrix



% -------------------------------------------------------------------------
% Parse input arguments
if nargin>=3, opts = varargin{1}; else opts = struct; end
rand('state',sum(100*clock));
costfcn = mgetopt(opts, 'costfcn', 'ls', 'instrset', {'ls','kl'});
Tau=mgetopt(opts,'Tau',[1 1]);
Phi=mgetopt(opts,'Phi',size(V,1)-Tau(1)+1);
const_WHD=mgetopt(opts,'const_WorH',[0 0 0]);
sparseD = mgetopt(opts,'sparseD', 0);
phiAlign=mgetopt(opts,'phiAling',0);
tauAlign=mgetopt(opts,'tauAling',0);
lambda = mgetopt(opts, 'lambda', 0);
maxiter = mgetopt(opts, 'maxiter', 100);
conv_criteria = mgetopt(opts, 'conv_criteria', 1e-6);
displaylevel = mgetopt(opts, 'displaylevel', 'iter', 'instrset', ...
    {'off','iter','final'});

% -------------------------------------------------------------------------
% Initialize W,H and D
mV=max(V(:));
W = mgetopt(opts, 'W', rand(Tau(1),d,Tau(2)));
H = mgetopt(opts, 'H', max(V(:))*rand(d,size(V,2),Phi));
D = mgetopt(opts, 'D', max(V(:))*rand(size(V,3),d));
if sparseD==0
   D=D/sqrt(sum(D(:).^2));
elseif sparseD==1
   D = normalizeW(D);
end
Q=mgetopt(opts,'Q',ones(size(V)));
V=V.*Q;
W = normalizeW(W);

% -------------------------------------------------------------------------
% Initialize Objective function
sst = sum((V(:)-mean(V(:))).^2);
Rec = ReconstructData(W, H, D, 1:d);
Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
switch costfcn
    case 'ls'
        sse = norm(V(:)-Rec(:),'fro')^2;
                cost_old = .5*sse + lambda*(sum(H(:)));
    case 'kl'
        ckl = sum(V(:).*log((V(:)+eps)./(Q(:).*Rec(:)+eps))-V(:)+Q(:).*Rec(:));
        cost_old = ckl + lambda*(sum(H(:)));
end
if sparseD==2
    cost_old=cost_old+lambda*sum(D(:));
end
delta_cost = 1;
iter = 0;
keepgoing = 1;
t=cputime;

% -------------------------------------------------------------------------
% Display information
dheader = sprintf('%12s | %12s | %12s | %12s | %12s','Iteration','Expl. var.','Cost func.','Delta costf.','iter. time');
dline = sprintf('-------------+--------------+--------------+--------------+--------------+');
if any(strcmp(displaylevel, {'final','iter'}))
    disp('Shift Invariant Sparse Coding');
    disp(['Strength of sparseness imposed ' num2str(lambda)]);
    disp(['Pct missing values ' num2str(length(find(Q==0))/prod(size(Q))*100)]);
    disp(['To stop algorithm press ctrl-C']);
    disp('');
end



% -------------------------------------------------------------------------
% Optimization loop
while keepgoing
    pause(0.00001);
    told=t;
    if mod(iter,10)==0
        if any(strcmp(displaylevel, {'iter'}))
            disp(dline); disp(dheader); disp(dline);
        end
    end

    % Align Phi and Tau values if specified
    if phiAlign | tauAlign
        if phiAlign
            [W,H] = alignPhi(W,H);
        end
        if tauAlign
            [W,H] = alignTau(W,H);
        end
        W = normalizeW(W);
        Rec = ReconstructData(W,H,D,1:d);
        Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
        Rec=Q.*Rec;
        switch costfcn
            case 'ls'
                cost_old = 0.5*norm(V(:)-Rec(:),'fro')^2+lambda*sum(H(:));
            case 'kl'
                cost_old = sum(V(:).*log((V(:)+eps)./(Rec(:)+eps))+Q(:))+lambda*sum(H(:));
        end
        if sparseD==2
             cost_old=cost_old+lambda*sum(D(:));
        end
    end
    

    % Update W, H and D
    switch costfcn
        case 'ls'
            [W,H,D,cost,sse,Rec] = ...
                sls_update(V,W,H,D,Tau,Phi,d,lambda,Rec,const_WHD,Q,sparseD);
        case 'kl'
            [W,H,D,cost,sse,Rec] = ...
                skl_update(V,W,H,D,Tau,Phi,d,lambda,Rec,const_WHD,Q,sparseD); 
    end

    delta_cost = cost_old - cost;
    cost_old=cost;
    iter=iter+1;

    % Display information
    if any(strcmp(displaylevel, {'iter'}))
        t=cputime;
        disp(sprintf('%12.0f | %12.4f | %12.6f | %11.2f | %11.2f', ...
            iter,(sst-sse)/sst,cost,delta_cost,t-told));
    end
    
    % Check if we should stop
    if delta_cost<cost*conv_criteria 
        if any(strcmp(displaylevel, {'iter','final'}))
           disp('SNMF2D has converged');
        end
        keepgoing = 0;
    end
    % Reached maximum number of iterations
    if iter>=maxiter
        if any(strcmp(displaylevel, {'iter','final'}))
            disp('Maximum number of iterations reached');
        end
        keepgoing=0; 
    end
end

% -------------------------------------------------------------------------
% Sparse Least squares update function
function [W,H,D,cost,sse,Rec] = ...
    sls_update(V,W,H,D,Tau,Phi,d,lambda,Rec,const_WHD,Q,sparseD)

% Update H
if ~const_WHD(2)
    Hx = zeros(size(H));
    Hy = zeros(size(H));
    for k=1:size(D,1)
        for d=1:size(D,2)
            Ht=conv2(D(k,d)*fliplr(flipud(squeeze(W(:,d,:))')),squeeze(V(:,:,k))');
            Hx(d,:,:)=squeeze(Hx(d,:,:))+Ht(size(W,3):size(W,3)+size(H,2)-1,size(W,1):size(H,3)+size(W,1)-1);
            Ht=conv2(D(k,d)*fliplr(flipud(squeeze(W(:,d,:))')),squeeze(Rec(:,:,k))');
            Hy(d,:,:)=squeeze(Hy(d,:,:))+Ht(size(W,3):size(W,3)+size(H,2)-1,size(W,1):size(H,3)+size(W,1)-1);
        end
    end    
    grad = Hx./(Hy+eps+lambda);
    H=H.*grad;
    Rec = ReconstructData(W,H,D,1:d); 
    Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
    Rec = Rec.*Q;
end

% Update W
if ~const_WHD(1)
    Wx = zeros(size(W));
    Wy = zeros(size(W));       
    for k=1:size(D,1)
        DH=repmat(D(k,:)',[1,size(H,2),size(H,3)]).*H;
        for phi = 1:Phi
            for tau = 1:Tau(2)
                Wx(1:min([Tau(1), size(V,1)-phi+1]), : , tau) = Wx(1:min([Tau(1), size(V,1)-phi+1]), : , tau) + ...
                    V(phi:min([size(V,1),phi+Tau(1)-1]), tau:end,k)*DH(:, 1:end-(tau-1), phi)';
                Wy(1:min([Tau(1), size(V,1)-phi+1]), : , tau) = Wy(1:min([Tau(1), size(V,1)-phi+1]), : , tau) + ...
                    Rec(phi:min([size(V,1),phi+Tau(1)-1]), tau:end,k)*DH(:, 1:end-(tau-1), phi)';     
            end
        end
    end    
    tx = sum(sum(Wy.*W,1),3);
    ty = sum(sum(Wx.*W,1),3);
    T=(repmat(tx,[size(W,1),1,size(W,3)])-repmat(ty,[size(W,1),1,size(W,3)])).*W;
    T1=zeros(size(W));
    I=find(T1>0);
    T1(I)=T(I);
    T2=abs(T-T1);
    Wx = Wx + T1;
    Wy = Wy + T2;
    grad = Wx./(Wy+eps);
    W = normalizeW(W.*(grad));
    Rec = ReconstructData(W,H,D,1:d); 
    Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
    Rec=Rec.*Q;
end

% Update D
if ~const_WHD(3)
    T = ones(1,size(D,2));
    V3 = matrizicing(V,3);
    Rec3 = matrizicing(Rec,3);
    for t=1:size(D,2)
        Rect2 = ReconstructData(W,H,T,t);
        Rect2=Rect2(1:size(V,1),1:size(V,2));
        Dx(:,t) = V3*Rect2(:);
        Dy(:,t) = Rec3*Rect2(:);
    end
    if sparseD==2
        grad = Dx./(Dy+lambda+eps);
    elseif sparseD==1        
        tx = sum(Dy.*D,1);
        ty = sum(Dx.*D,1);
        T = (repmat(tx,[size(D,1),1])-repmat(ty,[size(D,1),1])).*D;
        T1 = zeros(size(D));
        I = find(T>0);
        T1(I) = T(I);
        T2 = abs(T-T1);
        Dx = Dx + T1;
        Dy = Dy + T2;
        grad = Dx./(Dy+eps);
        D = normalizeW(D.*(grad));
    else
        tx = sum(sum(Dy.*D));
        ty = sum(sum(Dx.*D));
        txty=tx-ty;
        if txty>0
            Dx = Dx + txty*D;            
        else
            Dy = Dy - txty*D;
        end
        grad = Dx./(Dy+eps);
        D = D.*grad;
        D = D/sqrt(sum(D(:).^2));
    end
end
Rec = ReconstructData(W,H,D,1:d);
Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
Rec = Rec.*Q;
% Calculate objective function
sse = norm(V(:)-Rec(:),'fro')^2;
cost = .5*sse+lambda*sum(H(:));
if sparseD==2
    cost=cost+lambda*sum(D(:));
end


% -------------------------------------------------------------------------
% Sparse Kullback Leibler update function
function [W,H,D,cost,sse,Rec] = ...
    skl_update(V,W,H,D,Tau,Phi,d,lambda,Rec,const_WHD,Q,sparseD)



% Update H
if ~const_WHD(2)
    VR = V./(Rec+eps);
    Hx = zeros(size(H));
    Hy = zeros(size(H));
    for k=1:size(D,1)
        for d=1:size(D,2)
            Ht=conv2(D(k,d)*fliplr(flipud(squeeze(W(:,d,:))')),squeeze(VR(:,:,k))');
            Hx(d,:,:)=squeeze(Hx(d,:,:))+Ht(size(W,3):size(W,3)+size(H,2)-1,size(W,1):size(H,3)+size(W,1)-1);
            Ht=conv2(D(k,d)*fliplr(flipud(squeeze(W(:,d,:))')),squeeze(Q(:,:,k))');
            Hy(d,:,:)=squeeze(Hy(d,:,:))+Ht(size(W,3):size(W,3)+size(H,2)-1,size(W,1):size(H,3)+size(W,1)-1);
        end
    end
    grad = Hx./(Hy+eps+lambda);
    H = H.*grad;
    Rec = ReconstructData(W,H,D,1:d); 
    Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
    Rec = Rec.*Q;
end

% Update W
if ~const_WHD(1)
    VR = V./(Rec+eps);
    Wx = zeros(size(W));
    Wy = zeros(size(W));
    for k=1:size(D,1)
        DH=repmat(D(k,:)',[1,size(H,2),size(H,3)]).*H;
        for phi = 1:Phi
            for tau = 1:Tau(2)
                Wx(1:min([Tau(1), size(V,1)-phi+1]), : , tau) = Wx(1:min([Tau(1), size(V,1)-phi+1]), : , tau) + ...
                    VR(phi:min([size(V,1),phi+Tau(1)-1]), tau:end,k)*DH(:, 1:end-(tau-1), phi)';
                Wy(1:min([Tau(1), size(V,1)-phi+1]), : , tau) = Wy(1:min([Tau(1), size(V,1)-phi+1]), : , tau) + ...
                    Q(phi:min([size(V,1),phi+Tau(1)-1]), tau:end,k)*DH(:, 1:end-(tau-1), phi)';     
            end
        end
    end    
    tx = sum(sum(Wy.*W,1),3);
    ty = sum(sum(Wx.*W,1),3);
    Wx = Wx + repmat(tx,[size(W,1),1,size(W,3)]).*W;
    Wy = Wy + repmat(ty,[size(W,1),1,size(W,3)]).*W;
    W = normalizeW(W.*(Wx./(Wy+eps)));
    Rec = ReconstructData(W,H,D,1:d); 
    Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
    Rec = Rec.*Q;
end

% Update D
if ~const_WHD(3)
    T=ones(1,size(D,2));
    V3=matrizicing(V,3);
    Rec3=matrizicing(Rec,3);
    V3dRec3=V3./(Rec3+eps);
    O3=matrizicing(Q,3);
    for t=1:size(D,2)
        Rect2 = ReconstructData(W,H,T,t);
        Rect2=Rect2(1:size(V,1),1:size(V,2));
        Dx(:,t) = V3dRec3*Rect2(:);
        Dy(:,t) = O3*Rect2(:);
    end
   if sparseD==2
        grad = Dx./(Dy+lambda+eps);
    elseif sparseD==1        
        tx = sum(Dy.*D,1);
        ty = sum(Dx.*D,1);
        T = (repmat(tx,[size(D,1),1])-repmat(ty,[size(D,1),1])).*D;
        T1 = zeros(size(D));
        I = find(T>0);
        T1(I) = T(I);
        T2 = abs(T-T1);
        Dx = Dx + T1;
        Dy = Dy + T2;
        grad = Dx./(Dy+eps);
        D = normalizeW(D.*(grad));
    else
        tx = sum(sum(Dy.*D));
        ty = sum(sum(Dx.*D));
        txty=tx-ty;
        if txty>0
            Dx = Dx + txty*D;            
        else
            Dy = Dy - txty*D;
        end
        grad = Dx./(Dy+eps);
        D = D.*grad;
        D = D/sqrt(sum(D(:).^2));
    end
end
Rec = ReconstructData(W,H,D,1:d); 
Rec=Rec(1:size(V,1),1:size(V,2),1:size(V,3));
Rec = Rec.*Q;
% Calculate objective function
T=Rec-V;
ckl = sum(V(:).*log((V(:)+eps)./(Rec(:)+eps))+T(:));
cost = ckl + lambda*sum(H(:));
sse = norm(T(:),'fro')^2;
if sparseD
    cost=cost+lambda*sum(D(:));
end

% -------------------------------------------------------------------------
% Parser for optional arguments
function var = mgetopt(opts, varname, default, varargin)
if isfield(opts, varname)
    var = getfield(opts, varname); 
else
    var = default;
end
for narg = 1:2:nargin-4
    cmd = varargin{narg};
    arg = varargin{narg+1};
    switch cmd
        case 'instrset',
            if ~any(strcmp(arg, var))
                fprintf(['Wrong argument %s = ''%s'' - ', ...
                    'Using default : %s = ''%s''\n'], ...
                    varname, var, varname, default);
                var = default;
            end
        otherwise,
            error('Wrong option: %s.', cmd);
    end
end


% -------------------------------------------------------------------------
function W = normalizeW(W)
% Normalize W
N(1,:,1) = sqrt(sum(sum(W.^2,1),3));
W = W./repmat(N+eps,[size(W,1),1,size(W,3)]);

% -------------------------------------------------------------------------
function Y=matrizicing(X,n)

% The matrizicing operation, i.e.
% matrizicing(X,n) turns the multi-way array X into a matrix of dimensions:
% (size(X,n)) x (size(X,1)*...*size(X,n-1)*size(X,n+1)*...*size(X,N))
% where N=ndims(X)
%
% Input:
% X     Multi-way array
% n     Dimension along which matrizicing is performed
%
% Output:
% Y     The corresponding matriziced version of X

N=ndims(X);
Y=reshape(permute(X, [n 1:n-1 n+1:N]),size(X,n),prod(size(X))/size(X,n));

% -------------------------------------------------------------------------
% Scale W and H to ensure W and H keep within reasonable value range
function [W,H] = scaleWH(W,H)
Q(1,:,1) = sqrt(sum(sum(W.^2,1),3));
W = W./repmat(Q+eps,[size(W,1),1,size(W,3)]);
H=H.*repmat(Q',[1, size(H,2) size(H,3)]);


% -------------------------------------------------------------------------
function [W,H]=alignPhi(W,H)

for k = 1:size(W,2)
    md = squeeze(sum(W(:,k,:),3));
    q=sum((md/sum(md)).*[1:length(md)]');
    if 2*q>size(W,1)
        W(1:end-1,k,:)=W(2:end,k,:);
        H(k,:,2:end)=H(k,:,1:end-1);
    elseif 2*q<size(W,1)
        W(2:end,k,:)=W(1:end-1,k,:);
        H(k,:,1:end-1)=H(k,:,2:end);        
    end
end
    
% -------------------------------------------------------------------------
function [W,H]=alignTau(W,H)

for k = 1:size(W,2)
    md = squeeze(sum(W(:,k,:),1));
    q=sum((md/sum(md)).*[1:length(md)]');
    if 2*q>size(W,3)
        W(:,k,1:end-1)=W(:,k,2:end);
        H(k,2:end,:)=H(k,1:end-1,:);
    elseif 2*q<size(W,3)
        W(:,k,2:end)=W(:,k,1:end-1);
        H(k,1:end-1,:)=H(k,2:end,:);        
    end
end
